#include <stdlib.h>
#include "bintree.h"


#ifndef NULL
	#define NULL 0
#endif


/* Create a new binary search tree. The compare argument is the name of
 * the comparison function, which is called with two arguments that
 * point to two keys being compared. The function must return an integer
 * less than, equal to, or greater than zero to indicate if the first
 * key is to be considered less than, equal to, or greater than the second key.
 *
 */
TREE *new_tree( int (*compare)(const void *, const void *) )
{
	TREE *p = malloc( sizeof(TREE) );

	if ( p ) {
		p->root = NULL;
		p->compare = compare;
	}

	return p;
}

static NODE **find_node( NODE **n, void *key, int (*compare)() )
{
	int cmp;

	while ( *n != NULL && (cmp = compare(key, (*n)->key)) != 0 )
		n = &( cmp < 0 ? (*n)->lsub : (*n)->rsub );

	return n;
}

static NODE *mknode( void *key, void *data, NODE *lsub, NODE *rsub )
{
	NODE *n = malloc( sizeof(NODE) );

	if (n != NULL ) {
		 n->key = key;
		 n->data = data;
		 n->lsub = lsub;
		 n->rsub = rsub;
	}

	return n;
}


/* Create a new node with the given key and data, and insert it into
 * a tree. If a node already existed, it is not inserted and 0 is returned.
 * If a new node could not be created, -1 is returned. Otherwize,
 * on success, 1 is returned.
 *
 */
int insert_node( TREE *tree, void *key, void *data )
{
	NODE **n = find_node( &tree->root, key, tree->compare );

	if ( *n != NULL )
		return 0;

	if ( (*n = mknode( key, data, NULL, NULL )) == NULL )
		return -1;

	return 1;
}


static long traverse( NODE *n, long (*f)(NODE *p, void *aux), void *aux )
{
	if ( n == NULL )
		return 0;

	return traverse(n->lsub, f, aux) + f(n,aux) + traverse(n->rsub, f,aux);
}


/* Call the function f on all nodes in a tree and return the sum of the
 * return values from these calls.
 *
 * The tree is traversed in ascending order according to the
 * comparison function in effect.
 *
 */
long foreach_node( TREE *tree, long (*f)(NODE *p, void *aux), void *aux )
{
	return traverse( tree->root, f, aux );
}
