#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "filter.h"
#include "getline.h"
#include "bintree.h"
#include "iptools.h"
#include "die.h"


/************************************************************************
 *									*
 *		Read the input data and store it			*
 *									*
 ************************************************************************/

/* Each `Input' object represents one line read from the input source.
 * Each line contains an ip address range pluss some text.
 *
 */
typedef struct input {

	int octet[4];
	char *text;
	struct input *next;
} Input;


/* The Input objects are stored as linked lists, one list
 * per range-span per octet:
 */
static Input *inputs[4][256] = { {NULL} };


/* Create a new Input object and place it into the correct list in `inputs'.
 */
static void new_input( int octet[4], int oct, int range, char *text )
{
	Input *p = malloc_or_die( sizeof( Input ) );

	memcpy( p->octet, octet, sizeof( p->octet ) );
	p->text = text;

	p->next = inputs[ oct ][ range ];
	inputs[ oct ][ range ] = p;
}


void read_input( FILE *f, const char *filename,
		 char * (*rewrite_input)(char *) )
{
	char	*input;

	while (	input = getline( f ) ) {

		int	low[4] = {0};
		int	octet = 0;
		int	range;
		char	*p, *q;


		p = ( rewrite_input != NULL ) ? rewrite_input( input ) : input;


		/* Skip whitespace at the beginning of a line
		 * and skip blank lines
		 */
		for ( ; *p == ' ' || *p == '\t'; p++ );

		if ( *p == '\0' )
			continue;


		/* Parse the ip address range specification.
		 * The `octet' variable contains the number of octets read
		 */
		q = ip_get_octet( p, low + octet, &range );

		while ( q != p && ++octet < 4 ) {

			if ( *q == '.' )
				++q;

			if ( range > 0 )
				break;
			p = q;

			q = ip_get_octet( p, low + octet, &range );
		}

		if ( octet == 0 )
			die_syntax( input );


		/* `octet' from now on contains the octet number (0-3)
		 * where the expansion is to happen
		 */
		--octet;

		if ( range == 0 ) {

			int prefix = -1;

			q = ip_get_prefix( q, &prefix );

			if ( prefix > -1 ) {

				ip_normalize( low, prefix );

				octet = (prefix - 1) / 8;
				range = ip_naddr[ prefix ] - 1;
			}
		}

		if ( *q && *q != ' ' && *q != '\t' )
			die_syntax( input );

		new_input( low, octet, range, strdup_or_die( q ) );
	}

	if ( ! feof( f ) )
		die_io( filename );
}


/************************************************************************
 *									*
 *	 	     Generate and print the output			*
 *									*
 ************************************************************************/

static TREE *store = NULL;

static int compare( const void *a, const void *b )
{
	return strcmp( a, b );
}

/* Store the range+text.
 */
static void store_range( char *range, char *text )
{
	int err, x;
	char *t = strdup_or_die( range );

	err = (store == NULL && (store = new_tree( compare )) == NULL);

	if ( err || (x = insert_node( store, t, text )) == -1 )
		die_mem();    

	if ( x == 0 )
		free( t );
}

static long print_range( NODE *n, void *f )
{
	int x = fprintf( (FILE *)f, "%s%s\n", (char *)n->key, (char *)n->data);

	if ( x < 0 )
		die_io( "fprintf" );

	return 0;
}

static void print_ranges( FILE *f )
{
	if ( store != NULL )
		foreach_node( store, print_range, f );
}

void write_output( FILE *f, const char *filename, int add_dot,
		   char * (*rewrite_output)(char *) )
{
	int octet;
	int range;
	int i;

	/* All input has been read, parsed, normalized
	 * and stored. Now we make the
	 * shorter ranges override the wider ones and then print
	 * everything to the outout device.
	 *
	 */
	for ( octet =   3; octet >= 0; --octet )
	for ( range = 0; range <= 255; ++range ) {

		Input	*p = inputs[ octet ][ range ];
		char	*fmt;

		fmt = (octet == 0) ? "%d"	    :
		      (octet == 1) ? "%d.%d"	    :
		      (octet == 2) ? "%d.%d.%d"	    :
				     "%d.%d.%d.%d"  ;

		while ( p != NULL ) {

			for ( i = 0; i <= range; ++i ) {

				char buffer[128];

				sprintf(buffer, fmt, p->octet[0], p->octet[1],
						     p->octet[2], p->octet[3]);

				if ( add_dot && octet < 3 )
					strcat( buffer, "." );

				store_range( buffer, p->text );

				++p->octet[ octet ];
			}

			p = p->next;
		}
	}

	print_ranges( f );
}
