#include <ctype.h>
#include <stdlib.h>


/************************************************************************
 *									*
 *               Parse IP address range specifications			*
 *									*
 ************************************************************************/

/* Read an ip address octet, or octet range, from the string pointed to by `s'.
 *
 * Put the octet and range into the `octet' and `range' arguments.
 *
 * An octet is a decimal number in the range [0-255]. An octet range
 * consists of two octets separated by a dash (-). The difference between
 * the lower and higher octet is put into `range'.
 *
 * The return value is the position in `s' of the first character after the
 * octet/range just read. If a syntax error is detected, the return value
 * is set equal to the input `s' and `octet' and `range' are left untouched.
 *
 */
char *ip_get_octet( const char *s, int *octet, int *range )
{
	int	n;
	int	l, h;
	const unsigned char *p, *q;

	for ( n = 0, p = s; isdigit( *p ); ++n, ++p );

	if ( n == 0 || n > 3 || (l = h = atoi( s )) > 255 )
		return (char *)s;

	if ( *p == '-' ) {

		for ( n = 0, q = ++p; isdigit( *p ); ++n, ++p );

		if ( n == 0 || n > 3 || (h = atoi( q )) > 255 )
			return (char *)s;

		if ( h < l ) {
			int tmp = h; h = l; l = tmp;
		}
	}

	*octet = l;
	*range = h - l;

	return (char *)p;
}


/* Read an ip address prefix from the string pointed to by `s', and put it
 * into the `prefix' argument.
 *
 * An ip address prefix is a slash (/) followed by a number in the 
 * range [0-32]. The return value is the position in `s' of the first
 * character after the prefix just read. If a syntax error is detected,
 * `prefix' is left untouched and the return value is set equal to
 * the input `s'.
 *
 */
char *ip_get_prefix( const char *s, int *prefix )
{
	int n;
	int x;
	const unsigned char *p = s;

	if ( *p != '/' )
		return (char *)s;

	for ( n = 0, ++p; isdigit( *p ); ++n, ++p );

	if ( n == 0 || n > 2 || (x = atoi( s+1 )) > 32 )
		return (char *)s;

	*prefix = x;

	return (char *)p;
}


/* How many different octet values for the octet it ends in does a
 * prefix represent?
 *
 */
int ip_naddr[33] = { 256,
		     128, 64, 32, 16, 8, 4, 2, 1,
		     128, 64, 32, 16, 8, 4, 2, 1,
		     128, 64, 32, 16, 8, 4, 2, 1,
		     128, 64, 32, 16, 8, 4, 2, 1,
};

/* Normalize an IP address by making sure there are
 * only 0 bits in the `host portion'. If it needed to be normalized
 * return 1, otherwise, if it was already normalized, return 0.
 */
int ip_normalize( int octet[4], int prefix )
{
	int	n;    /* The octet that the prefix ends in     */
	int	ok;   /* Is the address and prefix consistant? */

	n = (prefix - 1) / 8;

	ok = octet[n] % ip_naddr[prefix] == 0;

	if ( !ok )
		octet[n] = (octet[n] / ip_naddr[prefix]) * ip_naddr[prefix];

	for ( ++n; n < 4; ++n )
		if ( octet[n] != 0 ) {
			ok = 0;
			octet[n] = 0;
		}

	return !ok;
}

int ip_consistant( int octet[4], int prefix )
{
	int	n;    /* The octet that the prefix ends in     */
	int	ok;   /* Is the address and prefix consistant? */

	n = (prefix - 1) / 8;

	ok = octet[n] % ip_naddr[prefix] == 0;

	for ( ++n; ok && n < 4; ++n ) {

		if ( octet[n] != 0 )
			ok = 0;
	}

	return ok;
}
